#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

double f,dt,t,pi2,amp,damp,scale;
long preblocks,blocks,postblocks,iok,isquare,hsquare,i,j,ii;
long dithering,payload,udchoice;

unsigned char pathname[512];
unsigned char leafname[64];
unsigned char outfilename[512];
unsigned char wtype,stype;
unsigned char command[512];
unsigned char filetag[64];

unsigned char appdirname[512];
unsigned char settingsfilename[512];
unsigned char userdatafilename[512];

unsigned char datablock[1152000];

long blocksize,icut,icut2,icut24,icut224,ibfactor,ok;


unsigned char ChunkID[5],Format[5],Subchunk1ID[5],Subchunk2ID[5];

long ChunkSize,Subchunk1Size,AudioFormat;
long NumChannels,SampleRate,ByteRate,BlockAlign;
long BitsPerSample,Subchunk2Size,BytesPerSample;

FILE* outfile;


void read_settings_file(void);

long even_test(long);
long ok_test(long);


long get_user_input(void);

void make_output_name(void);

void sinefix(void);
void squarefix(void);
void zero_block(void);
void save_block(void);

void make_sinewave(void);
void make_squarewave(void);
void make_trianglewave(void);
void make_impulsewave(void);
void make_userwave(void);
void load_values(double,long);
long quantise(double);
double dither(void);

void write_output_header(long);

void write4int(long);
void write2int(long);
void write4(unsigned char*);


int main(int arc,char*argv[])
{
  icut=32768; icut2=2*icut;  icut24=icut*256; icut224=2*icut24; 
  pi2=3.1415927*2.0;
  
  printf("  WAVGenerator Ver 1.05 8th June 2010\n");
  printf("  ===================================\n\n");
  printf("This lets you generate LPCM WAV files containing test waveforms.\n");
  printf("Please read the Help inside the application before first use\n");
  printf("and check you have the correct details in your Settings file!\n\n");
  
  sprintf(appdirname,"%s\0",argv[0]);
  ok=(long)strrchr(appdirname,'/')-(long)appdirname;
  appdirname[ok]=(char)0;
  printf("Application directory = %s\n",appdirname);
  sprintf(settingsfilename,"%s/%s\0",appdirname,"Settings\0");
  printf("Settings file at %s\n",settingsfilename);

  
  
  iok=get_user_input();
  read_settings_file();
  make_output_name();
  
  
  
  outfile=fopen(outfilename,"wd");
  
  write_output_header(payload);

  zero_block();
  j=0;
  while (j<preblocks)
  {
    save_block();
    j++;
  }
  
  if (wtype=='S') make_sinewave();
  if (wtype=='Q') make_squarewave();
  if (wtype=='T') make_trianglewave();
  if (wtype=='I') make_impulsewave();
  if (wtype=='U') make_userwave();
  
  
  
  j=0;
  while(j<blocks)
  {
    save_block();
    j++;
    printf("%3ld\n",j);
  }
  
  
  
  zero_block();
  j=0;
  while (j<postblocks)
  {
    save_block();
    j++;
  }
  
  fclose(outfile);
  printf("Track finished\n");
  
  
  
}

long get_user_input(void)
{
  long icheck,scheck;
  icheck=1;   /* defaults to sinusoid */
  f=400.0;    /* defaults to 400 Hz   */
  scheck=1; stype='M'; /* defaults to mono     */
  dithering=1; /* dithering enabled as default */
  
  sprintf(ChunkID,"RIFF\0");
  sprintf(Format,"WAVE\0");
  sprintf(Subchunk1ID,"fmt \0");
  sprintf(Subchunk2ID,"data\0");
  
  Subchunk1Size=16;
  AudioFormat=1;
  NumChannels=2;
  BlockAlign=4;
  
  printf("Choose bits per sample - Press\n");
  printf("  '1' for 16 bits per sample or\n");
  printf("  '2' for 24 bits per sample\n");
  printf("then press <enter> => ");
  scanf("%ld",&BitsPerSample);
  
  if (BitsPerSample==1)
  {
    BitsPerSample=16;
    BytesPerSample=2;
    scale=32767.0;
    ibfactor=4;
  }
  else
  {
    BitsPerSample=24;
    BytesPerSample=3;
    scale=256.0*32767.0;
    ibfactor=6;
  }
  
  printf("Choose sample rate - Press\n");
  printf("  '1' for  44.1k   '2' for  48k\n");
  printf("  '3' for  88.2k   '4' for  96k\n");
  printf("  '5' for 176.4k   '6' for 192k\n");
  printf("then press <enter> => ");
  scanf("%ld",&SampleRate);
  
  if(SampleRate<1)
  {
     SampleRate=1; printf("Out of range! Default to 44.1k\n");
  }
  if(SampleRate>6)
  {
     SampleRate=1; printf("Out of range! Default to 44.1k\n");
  }
  
  switch(SampleRate)
  {
    case 1:
      SampleRate=44100;   sprintf(filetag,"44k.wav\0");  udchoice=1;
      break;
    case 2:
      SampleRate=48000;   sprintf(filetag,"48k.wav\0");  udchoice=2;
      break;
    case 3:
      SampleRate=2*44100; sprintf(filetag,"88k.wav\0");  udchoice=1;
      break;
    case 4:
      SampleRate=2*48000; sprintf(filetag,"96k.wav\0");  udchoice=2;
      break;
    case 5:
      SampleRate=4*44100; sprintf(filetag,"176k.wav\0"); udchoice=1;
      break;
    case 6:
      SampleRate=4*48000; sprintf(filetag,"192k.wav\0"); udchoice=2;
  }
  
  if(udchoice==1)
  {
    sprintf(userdatafilename,"%s/UserData44\0",appdirname);
  }
  else
  {
    sprintf(userdatafilename,"%s/UserData48\0",appdirname);
  }
  
  printf("UserData from %s\n\n",userdatafilename);
  
  blocksize=SampleRate*BytesPerSample*2;
  dt=1.0/(double)SampleRate;
  ByteRate=blocksize;
  
  printf("Chosen values are:\n");
  printf(" Sample rate = %ld ksamples/sec per channel\n",SampleRate);
  printf("   Byte Rate = %ld bytes/sec  tag = %s\n",ByteRate,filetag);
  printf(" Sample interval %8.3lf microseconds\n",dt*1.0E6);
  
  printf("\n\n Choose S = Sinusoid / Q = Squarewave / T = Triangle\n        I = Impulse / U = User Defined\n");
  printf("Note Sinusoid and Triangle waves may be dithered, but other waveforms may not!\n");
  printf("  [type in character for chosen waveform and press <enter>.]\n\n => ");
  
  scanf("%c",&wtype);
  
  if ((long)wtype<14) scanf("%c",&wtype);
  
  if (wtype=='s') wtype='S';
  if (wtype=='q') wtype='Q';
  if (wtype=='t') wtype='T';
  if (wtype=='i') wtype='I';
  if (wtype=='u') wtype='U';
  
  
  if (wtype=='S') icheck=1;
  if (wtype=='Q') icheck=2;
  if (wtype=='T') icheck=3;
  if (wtype=='I') icheck=4;
  if (wtype=='U') icheck=0;
  
  dithering=0;
  if((icheck==1)||(icheck==3))
  {
    printf("Choose to dither or not. \n Enter 1 for dither and 0 for undithered =>  ");
    scanf("%ld",&dithering);
    if (dithering!=1) dithering=0;
    
    if (dithering==1)
    {
      printf("Dithering OK\n");
    }
    else
    {
      printf("Undithered output\n");
    }
    
  }
  
  
  printf("\n\n Choose M = Mono / L = Left / R = Right / A = Antiphase\n");
  
  scanf(" %c",&stype);
  
  if (stype=='m') stype='M';
  if (stype=='l') stype='L';
  if (stype=='r') stype='R';
  if (stype=='a') stype='A';
  
  if (stype=='L') scheck=2;
  if (stype=='R') scheck=3;
  if (stype=='A') scheck=4;
  
  
  printf("Chose %c  %ld\n\n",stype,scheck);
  
  printf("Enter amplitude [dbFS] => ");
  
  scanf("%lf",&damp);
  
  if (damp>0.0) damp=-damp;
  
  amp=damp/20.0;
  amp=pow(10.0,amp);
  printf("amplitude = %6.4lf wrt FS\n",amp);
  
  
  if(icheck>0)
  {
    printf("Enter frequency in Hz => ");
    scanf("%lf",&f);
    
    if (wtype=='S') sinefix();
    if (wtype=='T') squarefix();
    if (wtype=='I') squarefix();
    if (wtype=='Q') squarefix();
    
  }
    
  return icheck;
}



double dither(void)
{
  /* This procedure returns a 'random dither'
  levet as a real value in the range from -1.0
  to -1.0 with a triangular probability density */
  
  double answer;
  long dit;
  dit=rand()%128 + rand()%128;
  answer=(double)(dit-128);
  return answer/128.0;
}

void make_userwave(void)
{
  long iu,icheck,ulong,ifirst,ucount,istop;
  FILE* ufile;
  double unow;
  
  ifirst=0; iu=0; ucount=0; istop=0;
  
  
  ufile=fopen(userdatafilename,"rt");
  
  
  do
  {
    icheck=(long)fscanf(ufile,"%lf",&unow);
    
    if(icheck==1)
    {
      unow=unow*amp*scale;
      load_values(unow,iu);
      iu++;
    }
    else
    {
      rewind(ufile);
      ucount++;
      
      if(ifirst==0)
      {
        ifirst=1;
        ulong=iu;
      }
    }
    
    if(iu==SampleRate) istop=1;
    
  } while (istop==0);
  
  fclose(ufile);
  
  printf("User file read %ld samples per cycle\n",ulong);
  unow=(double)SampleRate/(double)ulong;
  printf(" Frequency %8.2lf Hz\n",unow);
  istop=SampleRate%ulong;
  
  if(istop==0)
  {
    printf("No cyclic remainder\n");
  }
  else
  {
    printf("Cyclic remainder %ld\n",istop);
  }
  
}


void make_sinewave(void)
{
  long ims;
  double answer;
  t=0.0; ims=0;
  do
  {
    answer=amp*scale*sin(f*t*pi2);
    if(dithering==1)
    {
      answer+=dither();
    }
    load_values(answer,ims);
    t+=dt;
    ims++;
  } while (ims<SampleRate);
  
  printf("sinewave created\n");
}


void make_squarewave(void)
{
  long ims,isq;
  double answer;
  
  
  answer=amp*scale;
  isq=0; ims=0;
  do
  {
    load_values(answer,ims);
    ims++;
    isq++;
    
    if (isq==hsquare)
    {
      isq=0;
      answer=(-answer);
    }
  } while (ims<SampleRate);
  
}


void make_impulsewave(void)
{
  long ims,iiq;
  double answer;
  
  answer=amp*scale;
  
  ims=0; iiq=0;
  load_values(-answer,ims);
  ims=1; iiq=1;
  do
  {
    if (iiq==hsquare)
    {
      load_values(answer,ims);
      answer=-answer;
      iiq=0;
    }
    else
    {
      load_values(0.0,ims);
    }
    
    ims++;
    iiq++;
  } while (ims<SampleRate);
  
}


void make_trianglewave(void)
{
  long ims,ist;
  double tristep,corner,vnow;
  
  corner=amp*scale;
  ims=0; ist=0;
  tristep=-2.0*corner/(double)hsquare;
  
  vnow=corner;
  do
  {
    if(dithering==1)
    {
      load_values(vnow+dither(),ims);
    }
    else
    {
      load_values(vnow,ims);
    }
    
    ims++;
    ist++;
    
    
    if (ist==hsquare)
    {
      corner=-corner;
      tristep=-tristep;
      vnow=corner;
      ist=0;
    }
    else
    {
      vnow+=tristep;
    }
  } while(ims<SampleRate);
}


long quantise(double rin)
{
  /* This proceedure takes a real value and
     returns a result as an integer which
     has been quantised into uniform steps. */
     
  long answer;
  
  if (rin>0.0)
  {
    answer=(long)(rin+0.5);
  }
  else
  {
    answer=(long)(rin-0.5);
  }
  return answer;
}


void load_values(double vin,long iv)
{
  long ib,lt,rt;
  unsigned char l1,l2,l3,r1,r2,r3;
  
  ib=ibfactor*iv;
  
  if(vin>scale)    vin=scale;
  if(vin<(-scale)) vin=(-scale);
  
  if(stype=='M')
  {
    lt=quantise(vin);
    rt=lt;
  }
  
  if(stype=='L')
  {
    lt=quantise(vin);
    rt=0;
  }
  
  if(stype=='R')
  {
    lt=0;
    rt=quantise(vin);
  }
  
  if(stype=='A')
  {
    lt=quantise(vin);
    rt=-lt;
  }
  
  if(BytesPerSample==2)
  {
    if(lt<0) lt+=2*icut;
    if(rt<0) rt+=2*icut;
    
    l1=(unsigned char)((lt>>8)&255);
    l2=(unsigned char)(lt&255);
    r1=(unsigned char)((rt>>8)&255);
    r2=(unsigned char)(rt&255);
    
    datablock[ib]=l2;
    datablock[ib+1]=l1;
    datablock[ib+2]=r2;
    datablock[ib+3]=r1;
  }
  else
  {
    
    if(lt<0) lt+=2*icut24;
    if(rt<0) rt+=2*icut24;
    
    l1=(unsigned char)((lt>>16)&255);
    l2=(unsigned char)((lt>>8)&255);
    l3=(unsigned char)(lt&255);
    
    r1=(unsigned char)((rt>>16)&255);
    r2=(unsigned char)((rt>>8)&255);
    r3=(unsigned char)(rt&255);
    
    datablock[ib]=l3;
    datablock[ib+1]=l2;
    datablock[ib+2]=l1;
    
    datablock[ib+3]=r3;
    datablock[ib+4]=r2;
    datablock[ib+5]=r1;
    
  }
}


void save_block(void)
{
  fwrite(datablock,sizeof(unsigned char),blocksize,outfile);
}


void zero_block(void)
{
  long iz;
  iz=0;
  do
  {
    datablock[iz]=(unsigned char)0;
    iz++;
  } while (iz<blocksize);
  
}



void sinefix(void)
{
  double fhfix;
  
  fhfix=(double)SampleRate;
  fhfix=fhfix*0.5;
  fhfix=fhfix-10.0;
  if (f<1.0) f=1.0;
  if (f>fhfix) f=fhfix;
  f=(double)(long)f;
  printf("Frequency (rounded to Hz) => %7.1lf Hz\n\n",f);
}

void squarefix(void)
{
  double ft,fdiff,fbest;
  long iseven,isok,ipass,ival;
  
  ival=1;
  fdiff=(double)SampleRate;
  
  
  do
  {
    iseven=even_test(ival);
    isok=ok_test(ival);
    ipass=iseven*isok;
    if(ipass==1)
    {
      ft=(double)SampleRate/(double)ival;
      if (fabs(ft-f)<fdiff)
      {
        fdiff=fabs(ft-f);
        fbest=ft;
        isquare=ival;
        hsquare=isquare/2;
      }
      
    }
    ival++;
  } while (ival<=SampleRate);
  
  
  
  f=fbest;
  
  
  printf("Correction for uniform rate to %ld samples per cycle\n",isquare);
  printf("Frequency %8.2lf Hz\n\n",f);
}


long even_test(long iet)
{
  long answer;
  if (iet%2==0)
  {
    answer=1;
  }
  else
  {
    answer=0;
  }
  return answer;
}


long ok_test(long iokt)
{
  long answer;
  if (SampleRate%iokt==0)
  {
    answer=1;
  }
  else
  {
    answer=0;
  }
  return answer;
}


void make_output_name(void)
{
  unsigned char ts[64];
  unsigned char du;
  long kpart,upart;
  
  if(dithering==1)
  {
    du='d';
  }
  else
  {
    du='u';
  }

  
  
  if (iok>0)
  {
    if (f>1000.0)
    {
      kpart=(long)(f/1000.0);
      upart=(long)f - kpart*1000;
      sprintf(ts,"%02ldk%03ldHz%03lddB\0",kpart,upart,(long)damp);
      printf("\na\n");
    }
    else
    {
      sprintf(ts,"%03ldHz%03lddB\0",(long)f,(long)damp);
      printf("\nb\n");
    }
    
    if(BytesPerSample==2)
    {
      sprintf(leafname,"%c%c%s%c\0",wtype,stype,ts,du);
    }
    else
    {
      sprintf(leafname,"%c%c%s%c24bit\0",wtype,stype,ts,du);
    }
    sprintf(outfilename,"%s/%s%s\0",pathname,leafname,filetag);
    printf("\nc\n");
    
  }
  else
  {
    if(BytesPerSample==2)
    {
      sprintf(leafname,"User%c%03lddB\0",stype,(long)damp);
    }
    else
    {
      sprintf(leafname,"User%c%03lddB24bit\0",stype,(long)damp);
    }
    sprintf(outfilename,"%s/%s%s\0",pathname,leafname,filetag);
    printf("\nd\n");
  }
  
  
  printf("Outfile file leafname => %s\n fullname => %s\n",leafname,outfilename);
}






void read_settings_file(void)
{
  /* Reads the settings file inside the application to allow
     some flexibility of the program settings */
  
  FILE* setfile;
  
  
  setfile=fopen(settingsfilename,"rt");
  

  
  fscanf(setfile,"%s",pathname);
  printf("Output to directory = %s\n",pathname); 
  
  fscanf(setfile,"%ld",&preblocks);
  fscanf(setfile,"%ld",&blocks);
  fscanf(setfile,"%ld",&postblocks);
  fclose(setfile);
  
  printf(" Set to produce track with test waveform of %ld sec duration\n preceeded by %ld sec silence and followed by %ld of silence.\n",blocks,preblocks,postblocks);
  
  payload=preblocks+blocks+postblocks;
  payload=payload*ByteRate;
  
  printf("Payload in Bytes = %ld\n",payload);
  
  
}


void write_output_header(long plo)
{
  
  printf("Payload to write = %ld\n",plo);
  write4(ChunkID);
  write4int(plo+36);
  write4(Format);
  write4(Subchunk1ID);
  write4int(Subchunk1Size);
  write2int(AudioFormat);
  write2int(NumChannels);
  write4int(SampleRate);
  write4int(ByteRate);
  write2int(BlockAlign);
  write2int(BitsPerSample);
  write4(Subchunk2ID);
  write4int(plo);
  
}

void write4int(long inv)
{
  unsigned char cop1,cop2,cop3,cop4;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  cop3=(unsigned char)((inv>>16)&255);
  cop4=(unsigned char)((inv>>24)&255);
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  fprintf(outfile,"%c",cop3);
  fprintf(outfile,"%c",cop4);
}


void write2int(long inv)
{
  unsigned char cop1,cop2;
  
  cop1=(unsigned char)(inv&255);
  cop2=(unsigned char)((inv>>8)&255);
  
  
  fprintf(outfile,"%c",cop1);
  fprintf(outfile,"%c",cop2);
  
}

void write4(unsigned char* outstring)
{
  long isp;
  
  isp=0;
  do
  {
    fprintf(outfile,"%c",outstring[isp]);
    isp++;
  } while (isp<4);
  
}


